import pytest
from starlette.routing import Route

from smyth.context import generate_context_data
from smyth.event import generate_api_gw_v2_event_data
from smyth.runner.strategy import first_warm
from smyth.server.app import SmythStarlette, create_app, lifespan
from smyth.server.endpoints import (
    invocation_endpoint,
    lambda_invoker_endpoint,
    status_endpoint,
)

pytestmark = pytest.mark.anyio


@pytest.fixture
def mock_get_config(mocker, config):
    return mocker.patch("smyth.server.app.get_config", return_value=config)


def test_create_app(mocker, mock_get_config):
    mock_smyth = mocker.Mock()
    mocker.patch("smyth.server.app.Smyth", autospec=True, return_value=mock_smyth)

    app = create_app()

    assert app.smyth == mock_smyth
    mock_smyth.add_handler.assert_has_calls(
        [
            mocker.call(
                name="order_handler",
                path=r"/test_handler",
                lambda_handler_path="tests.conftest.example_handler",
                timeout=None,
                event_data_function=generate_api_gw_v2_event_data,
                context_data_function=generate_context_data,
                log_level="DEBUG",
                concurrency=1,
                strategy_generator=first_warm,
                env_overrides={"TEST_ENV": "child", "ROOT_ENV": "root"},
            ),
            mocker.call(
                name="product_handler",
                path=r"/products/{path:path}",
                lambda_handler_path="tests.conftest.example_handler",
                timeout=None,
                event_data_function=generate_api_gw_v2_event_data,
                context_data_function=generate_context_data,
                log_level="DEBUG",
                concurrency=1,
                strategy_generator=first_warm,
                env_overrides={"TEST_ENV": "root", "ROOT_ENV": "root"},
            ),
        ]
    )


def test_smyth_starlette(mocker):
    mock_smyth = mocker.Mock()

    app = SmythStarlette(smyth=mock_smyth, smyth_path_prefix="/smyth")

    assert app.smyth == mock_smyth
    assert app.routes == [
        Route("/smyth/api/status", status_endpoint, methods=["GET", "HEAD"]),
        Route(
            "/2015-03-31/functions/{function:str}/invocations",
            invocation_endpoint,
            methods=["POST"],
        ),
        Route(
            "/{path:path}",
            lambda_invoker_endpoint,
            methods=["DELETE", "GET", "HEAD", "POST", "PUT", "OPTIONS"],
        ),
    ]


async def test_lifespan(mocker):
    mock_app = mocker.Mock()

    async with lifespan(mock_app):
        mock_app.smyth.start_runners.assert_called_once_with()

    mock_app.smyth.stop_runners.assert_called_once_with()

    mock_app = mocker.Mock()
    mock_app.smyth.start_runners.side_effect = Exception("Test")

    with pytest.raises(Exception):
        async with lifespan(mock_app):
            pass



import pytest

from smyth.exceptions import ProcessDefinitionNotFoundError
from smyth.runner.strategy import first_warm
from smyth.smyth import Smyth
from smyth.types import SmythHandlerState

pytestmark = pytest.mark.anyio


@pytest.fixture
def smyth(mock_event_data_function, mock_context_data_function):
    smyth = Smyth()
    smyth.add_handler(
        name="test_handler",
        path="/test_handler",
        lambda_handler_path="tests.conftest.example_handler",
        timeout=1,
        event_data_function=mock_event_data_function,
        context_data_function=mock_context_data_function,
        log_level="DEBUG",
        concurrency=1,
        strategy_generator=first_warm,
    )
    return smyth


def test_smyth_add_handler(smyth, mock_event_data_function, mock_context_data_function):
    handler = smyth.get_handler_for_name("test_handler")
    assert handler.name == "test_handler"
    assert handler.url_path.match("/test_handler")
    assert handler.lambda_handler_path == "tests.conftest.example_handler"
    assert handler.event_data_function == mock_event_data_function
    assert handler.context_data_function == mock_context_data_function
    assert handler.strategy_generator == first_warm
    assert handler.timeout == 1
    assert handler.log_level == "DEBUG"
    assert handler.concurrency == 1


def test_context_enter_exit(mocker):
    smyth = Smyth()
    mocker.patch.object(smyth, "start_runners")
    mocker.patch.object(smyth, "stop_runners")
    with smyth:
        pass
    assert smyth.start_runners.called
    assert smyth.stop_runners.called


def test_start_stop_runners(smyth):
    smyth.start_runners()
    assert smyth.processes["test_handler"][0].state == SmythHandlerState.COLD
    assert smyth.processes["test_handler"][0].task_counter == 0
    smyth.stop_runners()


def test_get_handler_for_request(smyth):
    handler = smyth.get_handler_for_request("/test_handler")
    assert handler.name == "test_handler"

    with pytest.raises(ProcessDefinitionNotFoundError):
        smyth.get_handler_for_request("/test_handler_2")


async def test_smyth_dispatch(
    smyth, mocker, mock_event_data_function, mock_context_data_function
):
    mock_asend = mocker.patch("smyth.runner.process.RunnerProcess.asend")
    mock_request = mocker.Mock()
    mock_request.method = "GET"
    mock_request.url.path = "/test_handler"

    with smyth:
        response = await smyth.dispatch(
            smyth.get_handler_for_name("test_handler"),
            mock_request,
        )

    assert mock_asend.await_args[0][0].type == "smyth.lambda.invoke"
    assert mock_asend.await_args[0][0].event == await mock_event_data_function()
    assert mock_asend.await_args[0][0].context == await mock_context_data_function()
    assert response == mock_asend.return_value


async def test_invoke(
    smyth, mocker, mock_event_data_function, mock_context_data_function
):
    mock_asend = mocker.patch("smyth.runner.process.RunnerProcess.asend")
    event_data = {"test": "data"}

    with smyth:
        response = await smyth.invoke(
            smyth.get_handler_for_name("test_handler"),
            event_data,
        )

    assert mock_asend.await_args[0][0].type == "smyth.lambda.invoke"
    assert mock_asend.await_args[0][0].event == event_data
    assert mock_asend.await_args[0][0].context == await mock_context_data_function()
    assert response == mock_asend.return_value



import pytest
from freezegun.api import FrozenDateTimeFactory
from pytest_mock import MockerFixture

from smyth.runner.fake_context import FakeLambdaContext


def test_fake_lambda_context(freezer: FrozenDateTimeFactory, mocker: MockerFixture):
    freezer.move_to("2024-12-20 00:00:00")
    expected_name = "Test Name Set In Test"
    environ = mocker.patch.dict("os.environ", clear=True)
    environ.update(
        {
            "AWS_LAMBDA_FUNCTION_NAME": expected_name,
        }
    )
    context = FakeLambdaContext()

    assert context.function_name == expected_name
    assert context.function_version == "$LATEST"
    assert context.invoked_function_arn == f"arn:aws:lambda:serverless:{expected_name}"
    assert context.memory_limit_in_mb == "128"
    assert context.aws_request_id == "1234567890"
    assert context.log_group_name == f"/aws/lambda/{expected_name}"
    assert context.log_stream_name == (
        "2024/12/20/[$LATEST]smyth_aws_lambda_log_stream_name"
    )


@pytest.mark.parametrize(
    (
        "name",
        "version",
        "timeout",
        "expected_name",
        "expected_version",
        "expected_timeout",
    ),
    [
        ("test", "test", 60, "test", "test", 60),
        ("test", "test", None, "test", "test", 6),
        ("test", None, 120, "test", "$LATEST", 120),
        (None, "test", 6, "Fake", "test", 6),
        (None, None, 6, "Fake", "$LATEST", 6),
    ],
)
def test_fake_lambda_context_with_params(
    name,
    version,
    timeout,
    expected_name,
    expected_version,
    expected_timeout,
    freezer: FrozenDateTimeFactory,
):
    freezer.move_to("2024-12-20 00:00:00")
    context = FakeLambdaContext(name=name, version=version, timeout=timeout)
    assert context.function_name == expected_name
    assert context.function_version == expected_version
    assert context._timeout == expected_timeout
    assert context.invoked_function_arn == f"arn:aws:lambda:serverless:{expected_name}"
    assert context.memory_limit_in_mb == "128"
    assert context.aws_request_id == "1234567890"
    assert context.log_group_name == f"/aws/lambda/{expected_name}"
    assert context.log_stream_name == (
        f"2024/12/20/[{expected_version}]smyth_aws_lambda_log_stream_name"
    )



import re
from unittest.mock import AsyncMock, Mock

import pytest

from smyth.config import Config
from smyth.types import RunnerProcessProtocol, SmythHandler, SmythHandlerState


@pytest.fixture(autouse=True)
def anyio_backend():
    return "asyncio", {"use_uvloop": True}


@pytest.fixture
def smyth_handler(
    mock_lambda_handler,
    mock_event_data_function,
    mock_context_data_function,
    mock_strategy_generator,
):
    return SmythHandler(
        name="test_handler",
        url_path=re.compile(r"/test_handler"),
        lambda_handler_path="tests.conftest.example_handler",
        event_data_function=mock_event_data_function,
        context_data_function=mock_context_data_function,
        strategy_generator=mock_strategy_generator,
        env_overrides={
            "TEST_ENV": "test",
        },
    )


@pytest.fixture
def config_toml_dict():
    return {
        "tool": {
            "smyth": {
                "host": "0.0.0.0",
                "port": 8080,
                "env": {"ROOT_ENV": "root", "TEST_ENV": "root"},
                "handlers": {
                    "order_handler": {
                        "handler_path": "tests.conftest.example_handler",
                        "url_path": "/test_handler",
                        "env": {"TEST_ENV": "child"},
                    },
                    "product_handler": {
                        "handler_path": "tests.conftest.example_handler",
                        "url_path": "/products/{path:path}",
                    },
                },
                "log_level": "INFO",
            }
        }
    }


@pytest.fixture
def config(config_toml_dict):
    return Config.from_dict(config_toml_dict["tool"]["smyth"])


def example_handler(event, context):
    return {"statusCode": 200, "body": "Hello, World!"}


@pytest.fixture
def mock_lambda_handler():
    return example_handler


@pytest.fixture
def mock_event_data_function():
    return AsyncMock(return_value={"key": "value"})


@pytest.fixture
def mock_context_data_function():
    return AsyncMock(return_value={"key": "value"})


@pytest.fixture
def mock_strategy_generator():
    return Mock()


@pytest.fixture
def mock_runner_process():
    mock = Mock(
        spec=RunnerProcessProtocol,
    )
    mock.name = "test_process"
    mock.task_counter = 0
    mock.last_used_timestamp = 0
    mock.state = SmythHandlerState.COLD
    return mock



import pytest
from starlette.testclient import TestClient

from smyth.exceptions import LambdaInvocationError, LambdaTimeoutError, SubprocessError
from smyth.server.app import SmythStarlette
from smyth.server.endpoints import dispatch
from smyth.types import LambdaResponse

pytestmark = pytest.mark.anyio


@pytest.fixture
def mock_smyth_dispatch(mocker):
    return mocker.AsyncMock()


@pytest.fixture
def mock_smyth(mocker, mock_smyth_dispatch):
    smyth = mocker.Mock()
    smyth.handlers = {
        "order_handler": mocker.Mock(name="order_handler"),
        "product_handler": mocker.Mock(name="product_handler"),
    }
    smyth.processes = {
        "order_handler": [
            mocker.Mock(name="process1", task_counter=0, state="cold"),
        ],
        "product_handler": [
            mocker.Mock(name="process2", task_counter=0, state="cold"),
            mocker.Mock(name="process3", task_counter=0, state="cold"),
        ],
    }
    smyth.dispatch = mock_smyth_dispatch
    return smyth


@pytest.fixture
def app(mock_smyth):
    return SmythStarlette(smyth=mock_smyth, smyth_path_prefix="/smyth")


@pytest.fixture
def test_client(app):
    return TestClient(app)


@pytest.mark.parametrize(
    ("side_effect", "expected_status_code", "expected_body"),
    [
        (None, 200, b"Hello, World!"),
        (LambdaInvocationError("Test error"), 502, b"Test error"),
        (LambdaTimeoutError("Test error"), 408, b"Lambda timeout"),
        (SubprocessError("Test error"), 500, b"Test error"),
    ],
)
async def test_dispatch(
    mocker,
    mock_smyth,
    mock_smyth_dispatch,
    side_effect,
    expected_status_code,
    expected_body,
):
    mock_request = mocker.Mock()
    mock_event_data_function = mocker.Mock()
    mock_smyth_dispatch.return_value = LambdaResponse(
        body="Hello, World!",
        status_code=200,
        headers={},
    )
    mock_smyth_dispatch.side_effect = side_effect
    response = await dispatch(
        smyth=mock_smyth,
        smyth_handler=mock_smyth.handlers["order_handler"],
        request=mock_request,
        event_data_function=mock_event_data_function,
    )
    assert response.status_code == expected_status_code
    assert response.body == expected_body


def test_status_endpoint(test_client):
    response = test_client.get("/smyth/api/status")
    assert response.status_code == 200
    assert response.json() == {
        "lambda handlers": {
            "order_handler": {
                "processes": [
                    {
                        "state": "cold",
                        "task_counter": 0,
                    }
                ]
            },
            "product_handler": {
                "processes": [
                    {
                        "state": "cold",
                        "task_counter": 0,
                    },
                    {
                        "state": "cold",
                        "task_counter": 0,
                    },
                ]
            },
        },
    }


def test_invocation_endpoint(test_client, mock_smyth, mock_smyth_dispatch):
    mock_smyth_dispatch.return_value = LambdaResponse(
        body="Hello, World!",
        status_code=200,
        headers={},
    )
    response = test_client.post(
        "/2015-03-31/functions/order_handler/invocations",
        json={"test": "test"},
    )
    assert response.status_code == 200
    assert response.text == "Hello, World!"



import os
import sys
from collections.abc import Awaitable, Callable, Iterator, MutableMapping
from dataclasses import dataclass
from enum import Enum
from re import Pattern
from time import strftime
from typing import Annotated, Any, Literal, Protocol, TypeAlias

from aws_lambda_powertools.utilities.typing import LambdaContext
from pydantic import BaseModel, Field
from starlette.requests import Request

LambdaEvent: TypeAlias = MutableMapping[str, Any]
EventData: TypeAlias = dict[str, Any]
EventDataCallable: TypeAlias = Callable[
    [Request, "SmythHandler", "RunnerProcessProtocol"], Awaitable[EventData]
]
ContextData: TypeAlias = dict[str, Any]
ContextDataCallable: TypeAlias = Callable[
    [Request | None, "SmythHandler", "RunnerProcessProtocol"], Awaitable[ContextData]
]
StrategyGenerator: TypeAlias = Callable[
    [str, dict[str, list["RunnerProcessProtocol"]]],
    Iterator["RunnerProcessProtocol"],
]
Environ: TypeAlias = dict[str, str]


class SmythHandlerState(str, Enum):
    COLD = "cold"
    WORKING = "working"
    WARM = "warm"


class RunnerInputMessage(BaseModel):
    type: str
    event: EventData | None = None
    context: ContextData | None = None


class LambdaResponse(BaseModel):
    status_code: int = Field(200, alias="statusCode")
    headers: dict[str, str] = {}
    body: str


class LambdaErrorResponse(BaseModel):
    type: str
    message: str
    stacktrace: str


class RunnerStatusMessage(BaseModel):
    type: Literal["smyth.lambda.status"]
    status: SmythHandlerState


class RunnerResponseMessage(BaseModel):
    type: Literal["smyth.lambda.response"]
    response: LambdaResponse


class RunnerErrorMessage(BaseModel):
    type: Literal["smyth.lambda.error"]
    error: LambdaErrorResponse


RunnerOutputMessage = Annotated[
    RunnerStatusMessage | RunnerResponseMessage | RunnerErrorMessage,
    Field(discriminator="type"),
]


LambdaHandler: TypeAlias = Callable[[LambdaEvent, LambdaContext], LambdaResponse]


class RunnerProcessProtocol(Protocol):
    name: str
    task_counter: int
    last_used_timestamp: float
    state: SmythHandlerState

    async def asend(self, data: RunnerInputMessage) -> LambdaResponse | None: ...

    def start(self) -> None: ...

    def stop(self) -> None: ...

    def send(self, data: RunnerInputMessage) -> LambdaResponse | None: ...

    def is_alive(self) -> bool: ...

    def terminate(self) -> None: ...

    def join(self) -> None: ...


@dataclass
class SmythHandler:
    name: str
    url_path: Pattern[str]
    lambda_handler_path: str
    event_data_function: EventDataCallable
    context_data_function: ContextDataCallable
    strategy_generator: StrategyGenerator
    timeout: float | None = None
    log_level: str = "INFO"
    concurrency: int = 1
    env_overrides: Environ | None = None

    def _get_env_value(self, key: str, default: str) -> str:
        """
        Helper method to retrieve an environment variable value with the following
        precedence:
        1. `self.env_overrides` if defined.
        2. `os.environ` if the key exists.
        3. The provided default value.
        """
        if self.env_overrides and key in self.env_overrides:
            return self.env_overrides[key]
        return os.environ.get(key, default)

    def get_environ(self) -> Environ:
        envs = {
            "_HANDLER": self._get_env_value("_HANDLER", self.lambda_handler_path),
            "AWS_ACCESS_KEY_ID": self._get_env_value(
                "AWS_ACCESS_KEY_ID", "000000000000"
            ),
            "AWS_SECRET_ACCESS_KEY": self._get_env_value(
                "AWS_SECRET_ACCESS_KEY", "test"
            ),
            "AWS_SESSION_TOKEN": self._get_env_value("AWS_SESSION_TOKEN", "test"),
            "AWS_DEFAULT_REGION": self._get_env_value(
                "AWS_DEFAULT_REGION", "eu-central-1"
            ),
            "AWS_REGION": self._get_env_value("AWS_REGION", "eu-central-1"),
            "AWS_EXECUTION_ENV": self._get_env_value(
                "AWS_EXECUTION_ENV",
                f"AWS_Lambda_python{sys.version_info.major}.{sys.version_info.minor}",
            ),
            "AWS_LAMBDA_FUNCTION_MEMORY_SIZE": self._get_env_value(
                "AWS_LAMBDA_FUNCTION_MEMORY_SIZE", "128"
            ),
            "AWS_LAMBDA_FUNCTION_NAME": self._get_env_value(
                "AWS_LAMBDA_FUNCTION_NAME", self.name
            ),
            "AWS_LAMBDA_FUNCTION_VERSION": self._get_env_value(
                "AWS_LAMBDA_FUNCTION_VERSION", "$LATEST"
            ),
            "AWS_LAMBDA_INITIALIZATION_TYPE": self._get_env_value(
                "AWS_LAMBDA_INITIALIZATION_TYPE", "on-demand"
            ),
            "AWS_LAMBDA_LOG_GROUP_NAME": self._get_env_value(
                "AWS_LAMBDA_LOG_GROUP_NAME", f"/aws/lambda/{self.name}"
            ),
            "AWS_LAMBDA_LOG_STREAM_NAME": self._get_env_value(
                "AWS_LAMBDA_LOG_STREAM_NAME",
                f"{strftime('%Y/%m/%d')}/[$LATEST]smyth_aws_lambda_log_stream_name",
            ),
            "AWS_LAMBDA_RUNTIME_API": self._get_env_value(
                "AWS_LAMBDA_RUNTIME_API", "127.0.0.1:9001"
            ),
            "AWS_XRAY_CONTEXT_MISSING": self._get_env_value(
                "AWS_XRAY_CONTEXT_MISSING", "LOG_ERROR"
            ),
            "AWS_XRAY_DAEMON_ADDRESS": self._get_env_value(
                "AWS_XRAY_DAEMON_ADDRESS", "127.0.0.1:2000"
            ),
        }

        return envs



import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from multiprocessing import set_start_method
from typing import Any

from starlette.applications import Starlette

from smyth.config import get_config, get_config_dict
from smyth.server.endpoints import (
    invocation_endpoint,
    lambda_invoker_endpoint,
    status_endpoint,
)
from smyth.smyth import Smyth
from smyth.utils import import_attribute

LOGGER = logging.getLogger(__name__)
set_start_method("spawn", force=True)


@asynccontextmanager
async def lifespan(app: "SmythStarlette") -> AsyncGenerator[None, None]:
    try:
        app.smyth.start_runners()
    except Exception as error:
        LOGGER.error("Error starting runners: %s", error)
        raise
    yield
    app.smyth.stop_runners()


class SmythStarlette(Starlette):
    smyth: Smyth

    def __init__(self, smyth: Smyth, smyth_path_prefix: str, *args: Any, **kwargs: Any):
        self.smyth = smyth
        kwargs["lifespan"] = lifespan
        super().__init__(*args, **kwargs)
        self.add_route(
            f"{smyth_path_prefix}/api/status", status_endpoint, methods=["GET"]
        )
        self.add_route(
            "/2015-03-31/functions/{function:str}/invocations",
            invocation_endpoint,
            methods=["POST"],
        )
        self.add_route(
            "/{path:path}",
            lambda_invoker_endpoint,
            methods=["GET", "POST", "PUT", "DELETE", "OPTIONS"],
        )


def create_app() -> SmythStarlette:
    LOGGER.debug("Creating app")
    config = get_config(get_config_dict())

    smyth = Smyth()

    for handler_name, handler_config in config.handlers.items():
        smyth.add_handler(
            name=handler_name,
            path=handler_config.url_path,
            lambda_handler_path=handler_config.handler_path,
            timeout=handler_config.timeout,
            event_data_function=import_attribute(
                handler_config.event_data_function_path
            ),
            context_data_function=import_attribute(
                handler_config.context_data_function_path
            ),
            log_level=handler_config.log_level,
            concurrency=handler_config.concurrency,
            strategy_generator=import_attribute(handler_config.strategy_generator_path),
            env_overrides=handler_config.get_env_overrides(config),
        )

    app = SmythStarlette(smyth=smyth, smyth_path_prefix=config.smyth_path_prefix)

    return app



import json
import os
from copy import deepcopy
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any

import toml

from smyth.exceptions import ConfigFileNotFoundError
from smyth.types import Environ


@dataclass
class HandlerConfig:
    handler_path: str
    url_path: str
    timeout: float | None = None
    event_data_function_path: str = "smyth.event.generate_api_gw_v2_event_data"
    context_data_function_path: str = "smyth.context.generate_context_data"
    log_level: str = "DEBUG"
    concurrency: int = 1
    strategy_generator_path: str = "smyth.runner.strategy.first_warm"
    env: Environ = field(default_factory=dict)

    def get_env_overrides(self, config: "Config") -> Environ:
        env = deepcopy(config.env)
        env.update(self.env)
        return env


@dataclass
class Config:
    host: str = "0.0.0.0"
    port: int = 8080
    handlers: dict[str, HandlerConfig] = field(default_factory=dict)
    log_level: str = "INFO"
    smyth_path_prefix: str = "/smyth"
    env: Environ = field(default_factory=dict)

    @classmethod
    def from_dict(cls, config_dict: dict[str, Any]) -> "Config":
        handler_data = config_dict.pop("handlers")
        handlers = {
            handler_name: HandlerConfig(**handler_config)
            for handler_name, handler_config in handler_data.items()
        }
        return cls(**config_dict, handlers=handlers)


def get_config_file_path(file_name: str = "pyproject.toml") -> Path:
    """Get config file path. If not found raise exception."""
    directory = Path.cwd()
    while not directory.joinpath(file_name).exists():
        if directory == directory.parent:
            raise ConfigFileNotFoundError(f"Config file {file_name} not found.")
        directory = directory.parent
    return directory.joinpath(file_name).resolve()


def get_config_dict(config_file_name: str | None = None) -> dict[str, Any]:
    """Get config dict."""
    if config_file_name:
        config_file_path = get_config_file_path(config_file_name)
    else:
        config_file_path = get_config_file_path()

    return toml.load(config_file_path)


def get_config(config_dict: dict[str, Any]) -> Config:
    """Get config."""
    if environ_config := os.environ.get("__SMYTH_CONFIG"):
        config_data = json.loads(environ_config)
        return Config.from_dict(config_data)
    return Config.from_dict(config_dict["tool"]["smyth"])


def serialize_config(config: Config) -> str:
    return json.dumps(asdict(config))



import os
import sys
from collections.abc import Callable
from time import strftime, time
from typing import Any

from aws_lambda_powertools.utilities.typing import LambdaContext


class FakeLambdaContext(LambdaContext):
    def __init__(
        self,
        name: str | None = None,
        version: str | None = None,
        timeout: int | None = None,
        **kwargs: Any,
    ):
        if name is None:
            name = os.environ.get("AWS_LAMBDA_FUNCTION_NAME", "Fake")
        self._name = name

        if version is None:
            version = os.environ.get("AWS_LAMBDA_FUNCTION_VERSION", "$LATEST")
        self._version = version

        self._created = time()

        if timeout is None:
            timeout = 6
        self._timeout = timeout

        for key, value in kwargs.items():
            setattr(self, key, value)

    def get_remaining_time_in_millis(self) -> int:  # type: ignore[override]
        return int(
            max(
                (self._timeout * 1000)
                - (int(round(time() * 1000)) - int(round(self._created * 1000))),
                0,
            )
        )

    @property
    def function_name(self) -> str:
        return self._name

    @property
    def function_version(self) -> str:
        return self._version

    @property
    def invoked_function_arn(self) -> str:
        return "arn:aws:lambda:serverless:" + self._name

    @property
    # This indeed is a string in the real context hence the ignore[override]
    def memory_limit_in_mb(self) -> str:  # type: ignore[override]
        return os.environ.get("AWS_LAMBDA_FUNCTION_MEMORY_SIZE", "128")

    @property
    def aws_request_id(self) -> str:
        return "1234567890"

    @property
    def log_group_name(self) -> str:
        return os.environ.get("AWS_LAMBDA_LOG_GROUP_NAME", f"/aws/lambda/{self._name}")

    @property
    def log_stream_name(self) -> str:
        return os.environ.get(
            "AWS_LAMBDA_LOG_STREAM_NAME",
            f"{strftime('%Y/%m/%d')}/[{self._version}]smyth_aws_lambda_log_stream_name",
        )

    @property
    def log(self) -> Callable[[str], int] | Any:
        return sys.stdout.write



import inspect
import logging
import logging.config
import os
import signal
import sys
import traceback
from collections.abc import Generator
from multiprocessing import Process, Queue, set_start_method
from queue import Empty
from time import time
from types import FrameType

from asgiref.sync import sync_to_async
from setproctitle import setproctitle

from smyth.exceptions import (
    LambdaHandlerLoadError,
    LambdaInvocationError,
    LambdaTimeoutError,
    SubprocessError,
)
from smyth.runner.fake_context import FakeLambdaContext
from smyth.types import (
    EventData,
    LambdaErrorResponse,
    LambdaHandler,
    LambdaResponse,
    RunnerErrorMessage,
    RunnerInputMessage,
    RunnerOutputMessage,
    RunnerResponseMessage,
    RunnerStatusMessage,
    SmythHandlerState,
)
from smyth.utils import get_logging_config, import_attribute

set_start_method("spawn", force=True)
LOGGER = logging.getLogger(__name__)


class RunnerProcess(Process):
    name: str
    task_counter: int
    last_used_timestamp: float
    state: SmythHandlerState

    def __init__(
        self,
        name: str,
        lambda_handler_path: str,
        log_level: str = "INFO",
        environ_override: dict[str, str] | None = None,
    ):
        self.name = name
        self.task_counter = 0
        self.last_used_timestamp = 0
        self.state = SmythHandlerState.COLD
        self.environ_override = environ_override

        self.environ: dict[str, str] = {
            "_HANDLER": lambda_handler_path,
        }

        if environ_override:
            self.environ.update(environ_override)

        self.input_queue: Queue[RunnerInputMessage] = Queue(maxsize=1)
        self.output_queue: Queue[RunnerOutputMessage] = Queue(maxsize=1)

        self.lambda_handler_path = lambda_handler_path
        self.log_level = log_level
        super().__init__(
            name=name,
        )

    def stop(self) -> None:
        self.input_queue.put(RunnerInputMessage(type="smyth.stop"))
        self.join()
        self.input_queue.close()
        self.output_queue.close()
        self.input_queue.join_thread()
        self.output_queue.join_thread()

    def send(self, data: RunnerInputMessage) -> LambdaResponse | None:
        LOGGER.debug("Sending data to process %s: %s", self.name, data)
        self.task_counter += 1
        self.last_used_timestamp = time()
        self.input_queue.put(data)

        while True:
            if not self.is_alive():
                LOGGER.error(
                    "Process is not alive, this should generally not happen. "
                    "Restart Smyth and check your configuration. "
                    "This often happens when the handler can't be loaded "
                    "(i.e. an exception is raised when importing the handler)."
                )
                raise SubprocessError("Process is not alive")
            try:
                message = self.output_queue.get(block=True, timeout=1)
            except Empty:
                continue
            except Exception as error:
                LOGGER.error("Error getting message from output queue: %s", error)
                return None

            LOGGER.debug("Received message from process %s: %s", self.name, message)

            if message.type == "smyth.lambda.status":
                self.state = message.status
            elif message.type == "smyth.lambda.response":
                self.state = SmythHandlerState.WARM
                return message.response
            elif message.type == "smyth.lambda.error":
                self.state = SmythHandlerState.WARM
                if message.error.type == "LambdaTimeoutError":
                    raise LambdaTimeoutError(message.error.message)
                else:
                    raise LambdaInvocationError(message.error.message)

    @sync_to_async(thread_sensitive=False)
    def asend(self, data: RunnerInputMessage) -> LambdaResponse | None:
        return self.send(data)

    # Backend

    def run(self) -> None:
        setproctitle(f"smyth:{self.name}")
        logging.config.dictConfig(get_logging_config(self.log_level))
        os.environ.update(self.environ)
        self.lambda_invoker__()

    def get_message__(self) -> Generator[RunnerInputMessage, None, None]:
        while True:
            try:
                message = self.input_queue.get(block=True, timeout=1)
            except KeyboardInterrupt:
                LOGGER.debug("Stopping process")
                return
            except Empty:
                continue
            else:
                LOGGER.debug("Received message: %s", message)
                if message.type == "smyth.stop":
                    LOGGER.debug("Stopping process")
                    return
                yield message

    def get_event__(self, message: RunnerInputMessage) -> EventData:
        if message.event is None:
            raise LambdaInvocationError("No event data provided")
        return message.event

    def get_context__(self, message: RunnerInputMessage) -> FakeLambdaContext:
        if message.context is None:
            raise LambdaInvocationError("No context data provided")
        return FakeLambdaContext(**message.context)

    def import_handler__(
        self, lambda_handler_path: str, event: EventData, context: FakeLambdaContext
    ) -> LambdaHandler:
        LOGGER.info("Starting cold, importing '%s'", lambda_handler_path)
        try:
            handler: LambdaHandler = import_attribute(lambda_handler_path)
        except ImportError as error:
            raise LambdaHandlerLoadError(
                f"Error importing handler: {error}, module not found"
            ) from error
        except AttributeError as error:
            raise LambdaHandlerLoadError(
                f"Error importing handler: {error}, attribute in module not found"
            ) from error

        sig = inspect.signature(handler)
        try:
            sig.bind(event, context)
        except TypeError:
            LOGGER.warning(
                "Handler signature does not match event and context, "
                "using `event` and `context` as parameters."
            )
        return handler

    def set_status__(self, status: SmythHandlerState) -> None:
        self.output_queue.put(
            RunnerStatusMessage(type="smyth.lambda.status", status=status)
        )

    @staticmethod
    def timeout_handler__(signum: int, frame: FrameType | None) -> None:
        raise LambdaTimeoutError("Lambda timeout")

    def lambda_invoker__(self) -> None:
        sys.stdin = open("/dev/stdin")
        lambda_handler: LambdaHandler | None = None
        self.set_status__(SmythHandlerState.COLD)

        for message in self.get_message__():
            if message.type != "smyth.lambda.invoke":
                LOGGER.error("Invalid message type: %s", message.type)
                continue

            event = self.get_event__(message)
            context = self.get_context__(message)

            if not lambda_handler:
                lambda_handler = self.import_handler__(
                    self.lambda_handler_path,
                    event,
                    context,
                )
                self.set_status__(SmythHandlerState.WARM)

            signal.signal(signal.SIGALRM, self.timeout_handler__)
            signal.alarm(int(context._timeout))
            self.set_status__(SmythHandlerState.WORKING)
            try:
                response = lambda_handler(event, context)
            except Exception as error:
                LOGGER.exception(
                    "Error invoking lambda: %s",
                    error,
                    extra={"log_setting": "console_full_width"},
                )
                self.output_queue.put(
                    RunnerErrorMessage(
                        type="smyth.lambda.error",
                        error=LambdaErrorResponse(
                            type=type(error).__name__,
                            message=str(error),
                            stacktrace=traceback.format_exc(),
                        ),
                    )
                )
            else:
                self.output_queue.put(
                    RunnerResponseMessage(
                        type="smyth.lambda.response",
                        response=response,
                    )
                )
            finally:
                signal.alarm(0)
